#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/io.hpp>

#include "uquadprog.hpp"

#include "libmquad.h"

using namespace boost::numeric::ublas;

static matrix<double> cross_matrix(vector<double> V){
	matrix<double> M(3,3);
	M(0,0)=  0.0; M(0,1)=-V(2); M(0,2)= V(1);
	M(1,0)= V(2); M(1,1)=  0.0; M(1,2)=-V(0);
	M(2,0)=-V(1); M(2,1)= V(0); M(2,2)= 0.0;
	return M;
}

#define MASS	params[0]
#define HEIGHT	params[1]
#define	DX		params[2]	// half length of foot
#define	DY		params[3]	// half width of foot
#define MU		params[4]	// coefficient of friction

void libmquad(double *X, double *ddP, double *Ftask, double *Mtask, double *params, int contact_state, double *w, double *F){
	int i;
	double f[1];
	static matrix<double> G(12,12); 
	static vector<double>g0(12);
	static vector<double>x(12);

	static vector<double>p(3);
	static vector<double>pR(3);
	static vector<double>pL(3);
	static vector<double>ddp_des(3);

	matrix<double> CE(12,6);
	vector<double>ce0(6);
    
    static matrix<double>Aeq(6,12);
    static vector<double>beq(6);
    
    static matrix<double>A(18,12);
    static vector<double>b(18);    
    
    static int libmquad_initialized = 0;
   
	
    CE.clear();
    ce0.clear();

    switch(contact_state){
        case 0:	// DSP
            // nothing
            CE.resize(12,0,false);
            ce0.resize(0);
            break;
        case 1:	// LSP
            for(i=0;i<6;i++)	CE(6+i,i) = 1.0;
            break;
        case 2:	// RSP
            for(i=0;i<6;i++)	CE(i,i) = 1.0;
            break;
		case 3: // AIR
			for(i=0;i<12;i++)	F[i] = 0.0;
			return;
			break;
    }

    //if (!libmquad_initialized){

       
        Aeq.clear();

        for(i=0;i<6;i++){
            Aeq(i,i) = 1.0;
            Aeq(i,6+i) = 1.0;
        }
        
        A.clear();

        // COP constraints
        A( 0,2) =  DX; A( 0,4) =  1.0;	b( 0) = 0.0;		// DX*FLZ + MLY >= 0
        A( 1,2) =  DX; A( 1,4) = -1.0;	b( 1) = 0.0;		// DX*FLZ - MLY >= 0
        A( 2,2) =  DY; A( 2,3) =  1.0;	b( 2) = 0.0;		// DY*FLZ + MLX >= 0
        A( 3,2) =  DY; A( 3,3) = -1.0;	b( 3) = 0.0;		// DY*FLZ - MLX >= 0

        A( 4,8) =  DX; A( 4,10) =  1.0;	b( 4) = 0.0;		// DX*FRZ + MRY >= 0
        A( 5,8) =  DX; A( 5,10) = -1.0;	b( 5) = 0.0;		// DX*FRZ - MRY >= 0
        A( 6,8) =  DY; A( 6,9)  =  1.0;	b( 6) = 0.0;		// DY*FRZ + MRX >= 0
        A( 7,8) =  DY; A( 7,9)  = -1.0;	b( 7) = 0.0;		// DY*FRZ - MRX >= 0

        A( 8,2) =  1;					b( 8) = 0.0;		// FLZ >= 0
        A( 9,8) =  1;					b( 9) = 0.0;		// FRZ >= 0
        
        libmquad_initialized = 1;
    //}


	p(0) = X[0];
	p(1) = X[1];
	p(2) = HEIGHT;

	pL(0) = X[4];
	pL(1) = X[5];
	pL(2) = 0.0;

	pR(0) = X[6];
	pR(1) = X[7];
	pR(2) = 0.0;

	ddp_des(0) = ddP[0];
	ddp_des(1) = ddP[1];
	ddp_des(2) = ddP[2];;

	//===== Set Aeq*x - beq = 0, corresponds to the dynamic equations

	matrix_range<matrix<double> > Aeq1(Aeq, range (3, 6), range (0, 3));
	matrix_range<matrix<double> > Aeq2(Aeq, range (3, 6), range (6, 9));
	
	Aeq1 = cross_matrix(pL-p);
	Aeq2 = cross_matrix(pR-p);

	beq(0) = MASS*ddp_des(0) + Ftask[0];
	beq(1) = MASS*ddp_des(1) + Ftask[1];
	beq(2) = MASS*(ddp_des(2) + 9.81) + Ftask[2];
	beq(3) = Mtask[0];
	beq(4) = Mtask[1];
	beq(5) = Mtask[2];

    // make moment balance less important
    matrix_range<matrix<double> > AeqM(Aeq, range (3, 6), range (0, 12));
    AeqM = 0.3*AeqM;
    beq(3) = 0.3*beq(3);
    beq(4) = 0.3*beq(4);
    beq(5) = 0.3*beq(5);  

	//===== Set A*x - b >= 0, corresponding to the contact constraints
	
	// Friction constraints
	A(10,2) =  MU; A(10,0) =  1.0;	b(10) = 0.0;		// MU*FLZ + FLX >= 0
	A(11,2) =  MU; A(11,0) = -1.0;	b(11) = 0.0;		// MU*FLZ - FLX >= 0
	A(12,2) =  MU; A(12,1) =  1.0;	b(12) = 0.0;		// MU*FLZ + FLY >= 0
	A(13,2) =  MU; A(13,1) = -1.0;	b(13) = 0.0;		// MU*FLZ - FLY >= 0

	A(14,8) =  MU; A(14,6) =  1.0;	b(14) = 0.0;		// MU*FLZ + FLX >= 0
	A(15,8) =  MU; A(15,6) = -1.0;	b(15) = 0.0;		// MU*FLZ - FLX >= 0
	A(16,8) =  MU; A(16,7) =  1.0;	b(16) = 0.0;		// MU*FLZ + FLY >= 0
	A(17,8) =  MU; A(17,7) = -1.0;	b(17) = 0.0;		// MU*FLZ - FLY >= 0

	//===== Construct the quadratic objective function 
	// 0.5*(Aeq*x-beq)^T*(Aeq*x-beq) + 0.5*x^T*W*x
	//
	//	G = Aeq^T*Aeq + W
	//	g0 = -beq^T*Aeq

	G = prod(trans(Aeq),Aeq);
	for(i=0;i<12;i++)	G(i,i) += w[i]*1.0e-4;

	g0 = -prod(trans(beq),Aeq);
    
    f[0] = solve_quadprog(G, g0,  CE, ce0,  trans(A), b, x);

	for(i=0;i<12;i++)	F[i] = x(i); 

	
	if (F[2] > 1.1*MASS*9.81 ){
		F[2] = 1.1*MASS*9.81;
		F[8] = 0.0;
	}
	if (F[2] < 0)	F[2] = 0.0;

	if (F[8] > 1.1*MASS*9.81 ){
		F[8] = 1.1*MASS*9.81;
		F[2] = 0.0;
	}
	if (F[8] < 0)	F[8] = 0.0;

	if (F[4] >  DX*F[2])	F[4] =  DX*F[2];
	if (F[4] < -DX*F[2])	F[4] = -DX*F[2];
	if (F[3] >  DY*F[2])	F[3] =  DY*F[2];
	if (F[3] < -DY*F[2])	F[3] = -DY*F[2];

	if (F[10]>  DX*F[8])	F[10]=  DX*F[8];
	if (F[10]< -DX*F[8])	F[10]= -DX*F[8];
	if (F[9] >  DY*F[8])	F[9] =  DY*F[8];
	if (F[9] < -DY*F[8])	F[9] = -DY*F[8];

	for(i=0;i<12;i++)	x(i) = F[i]; 
	
	vector<double>z(6);
	z = prod(Aeq,x);

	Ftask[0] = z(0);
	Ftask[1] = z(1);
	Ftask[2] = z(2);
	Mtask[0] = z(3);
	Mtask[1] = z(4);
	Mtask[2] = z(5);


}

//
//#ifdef MATLAB
//void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]){
//	/*	INPUT: 
//     *      X - state [x,y,dx,dy,plx,ply,prx,pry]
//     *      ddX - desired acceleration [ddx,ddy]
//			P - params [mass,L,dx,dy]
//			w - solution weights (1 x 12)
//		OUTPUT:		
//            F - forces
//	*/
//#else
//void main(void){
//#endif
//	int i;
//#ifdef MATLAB
//    double *X, *ddX, *params, *F, *w;
//    
//    if(nrhs!=4) mexErrMsgTxt("libmquad: Wrong number of input arguments");
//    
//    if (mxGetM(prhs[0])!=8 && mxGetN(prhs[0])!=1){
//        mexPrintf("should be 8 x 1\n");
//        mexErrMsgTxt("libmquad: Wrong size of input argument 1"); 
//    }
//	if (mxGetM(prhs[1])!=2){
//        mexPrintf("should be 2 x 1\n");
//        mexErrMsgTxt("libmquad: Wrong size of input argument 2"); 
//    }    
//	if (mxGetM(prhs[2])!=1 || mxGetN(prhs[2])!=5){
//        mexPrintf("should be 1 x 5\n");
//        mexErrMsgTxt("libmquad: Wrong size of input argument 3"); 
//    }    
//	if (mxGetM(prhs[3])!=1 || mxGetN(prhs[3])!=12){
//        mexPrintf("should be 1 x 12\n");
//        mexErrMsgTxt("libmquad: Wrong size of input argument 3"); 
//    }     
//    X = mxGetPr(prhs[0]);
//    ddX = mxGetPr(prhs[1]);    
//    params = mxGetPr(prhs[2]); 
//	w = mxGetPr(prhs[3]);
//
//    plhs[0] = mxCreateDoubleMatrix(12,1,mxREAL);
//	F = mxGetPr(plhs[0]);
//#else
//	double X[8]={0,0,0,0,0,0.1,0.0,-0.1};
//	double ddX[2] = {0.0,-1.0};
//	double params[5]={95,0.9,0.1,0.05,0.5};
//	double w[12];
//	for(i=0;i<12;i++)	w[i] = 1.0e-4;
//	double F[12];
//#endif
//	
//	libmquad(X,ddX,params,w,F);
//    
//}
